@@ -12,35 +12,44 @@ from datetime import timedelta
 from collections import defaultdict
 from pywss.headers import *
 from pywss.statuscode import *
+from pywss.middleware import *
 from pywss.websocket import WebSocketUpgrade, WebSocketContextWrap
 from pywss.testing import HttpTestRequest, HttpTestResponse
-from pywss.static import NewStaticHandler
 from pywss.closing import Closing
 from pywss.routing import Route
 from pywss.openapi import openapi_ui_template, merge_dict, parameters_filter
 
-__version__ = '0.1.13'
+__version__ = '0.1.15'
 
 
 class Context:
     _handler_index = 0
     _flush_header = False
 
-    def __init__(self, app, fd, address, log, rfd, method, path, paths, version, headers, route, handlers):
+    def __init__(
+            self, app, log,
+            fd, rfd, address,
+            http_method, http_path, http_version,
+            route, route_keys, headers,
+            register_route, register_handlers
+    ):
         self.app: 'App' = app
         self.log: loggus.Entry = log
         self.fd: socket.socket = fd
         self.rfd: BufferedReader = rfd
-        self.method: str = method
-        self.version: str = version
-        self.headers: dict = headers
-        self.cookies: dict = parse_cookies(headers)
-        self.path: str = path
-        self.paths: dict = paths
-        self.params: dict = parse_params(path)
-        self.route: str = route
-        self._handlers: list = handlers
         self.address: tuple = address
+        self.method: str = http_method
+        self.path: str = http_path  # /route?key=value
+        self.version: str = http_version
+        self.route: str = route  # /route
+        self.route_keys: dict = route_keys
+        self.headers: dict = headers
+        self._route: str = register_route
+        self._handlers: list = register_handlers
+
+        self.cookies: dict = parse_cookies(headers)
+        self.params: dict = parse_params(http_path)
+        self._stream: bool = False  # make True when using self.stream(), then self.body() will always empty
         self.data: Data = Data()  # data save for user
 
         self.content_length: int = int(headers.get("Content-Length", 0))
@@ -96,10 +105,39 @@ class Context:
         return resp
 
     def body(self) -> bytes:
+        if self._stream:
+            return self.content  # should be empty
         if not self.content and self.content_length:
             self.content = self.rfd.read(self.content_length)
+        if not self.content and self.headers.get("Transfer-Encoding", "").lower() == "chunked":
+            size = int(self.rfd.readline(), 16)
+            while size > 0:
+                self.content += self.rfd.read(size)
+                assert self.rfd.read(2) == b"\r\n"
+                size = int(self.rfd.readline(), 16)
+            assert self.rfd.read(2) == b"\r\n"
         return self.content
 
+    def stream(self, size=65536):
+        if self._stream:
+            return self.content  # should be empty
+        if not self.content and self.content_length:
+            cl = self.content_length
+            while cl > 0:
+                rl = min(size, cl)
+                yield self.rfd.read(rl)
+                cl -= rl
+            self._stream = True
+            return
+        if not self.content and self.headers.get("Transfer-Encoding", "").lower() == "chunked":
+            size = int(self.rfd.readline(), 16)
+            while size > 0:
+                yield self.rfd.read(size)
+                assert self.rfd.read(2) == b"\r\n"
+                size = int(self.rfd.readline(), 16)
+            assert self.rfd.read(2) == b"\r\n"
+            self._stream = True
+
     def set_header(self, k, v) -> None:
         header = []
         for key in k.split("-"):
@@ -235,19 +273,21 @@ class Context:
 class App:
 
     def __init__(self, base_route="", base_handlers=None):
-        self.base_route = f"/{base_route.strip().strip('/')}" if base_route else base_route
-        self.base_handlers = list(base_handlers) if base_handlers else []
-        self.head_match_routes = []
-        self.parse_match_routes = []
-        self.full_match_routes = {}
-        self.log = loggus.GetLogger()
-        self.openapi_data = {
+        self.base_route: str = f"/{base_route.strip().strip('/')}" if base_route else base_route
+        self.base_handlers: list = list(base_handlers) if base_handlers else []
+        self.head_match_routes: list = []
+        self.parse_match_routes: list = []
+        self.full_match_routes: dict = {}
+        self.openapi_data: dict = {
             "paths": defaultdict(dict),
         }
-        self.running = False
-        Closing.add_close(self.close)
+        self.running: bool = False
+        self.data: Data = Data()
+        self.log = loggus.GetLogger()
 
     def register(self, method, route, handlers) -> None:
+        if len(handlers) < 1:
+            raise Exception("not found handlers")
         route = f"/{route.strip().strip('/')}" if route else route
         route = f"{self.base_route}{route}"
         route = f"/{route.strip().strip('/')}"
@@ -366,6 +406,17 @@ class App:
         self.patch(route, *handlers)
         self.options(route, *handlers)
 
+    def view(self, route, *handlers):
+        if len(handlers) < 1:
+            raise Exception("not found handlers")
+        view = handlers[-1]
+        handlers = list(handlers[:-1])
+        if hasattr(view, "use"):
+            handlers += list(getattr(view, "use")() or [])
+        for method in ("get", "post", "head", "put", "delete", "patch", "options", "any"):
+            if hasattr(view, f"http_{method}"):
+                getattr(self, method)(route, *handlers, getattr(view, f"http_{method}"))
+
     def openapi(
             self,
             title="OpenAPI",
@@ -391,55 +442,64 @@ class App:
             openapi_ui_css_url,
         )))
 
-    def _(self, request: socket.socket, address: tuple) -> None:
+    def handler_request(self, request: socket.socket, address: tuple) -> None:
         log = self.log
         try:
             rfd = request.makefile("rb", -1)
             while True:
-                method, path, version, err = parse_request_line(rfd)
+                http_method, http_path, http_version, err = parse_request_line(rfd)
                 if err:
                     request.sendall(b"HTTP/1.1 400 BadRequest\r\n")
                     return
-                route = f"{method.upper()}/{path.split('?', 1)[0].strip('/')}"
-                log = log.update(route=route)
-                hes, err = parse_headers(rfd)
+                route = register_route = f"/{http_path.split('?', 1)[0].lstrip('/')}"
+                route_keys = dict()
+                method_route = f"{http_method.upper()}{route.rstrip('/')}"
+                log = log.update(route=method_route)
+                http_headers, err = parse_headers(rfd)
                 if err:
                     request.sendall(b"HTTP/1.1 400 BadRequest\r\n")
                     log.error(err)
                     return
                 # check keep alive
-                if hes.get("Connection", "").lower() == "close":
+                if http_headers.get("Connection", "").lower() == "close":
                     break
                 # full match
-                paths = {}
-                handlers = self.full_match_routes.get(route, None)
+                handlers = self.full_match_routes.get(method_route, None)
                 # parser match
                 if not handlers:
                     for r, v in self.parse_match_routes:
-                        fix, ps = r.match(route)
+                        fix, ps = r.match(method_route)
                         if fix:
-                            route = r.route
+                            register_route = r.route
                             handlers = v
-                            paths = ps
+                            route_keys = ps
                             break
                 # head match
                 if not handlers:
                     for r, v in self.head_match_routes:
-                        if route.startswith(r):
-                            route = r
+                        if method_route.startswith(r):
+                            register_route = r
                             handlers = v
                             break
                 if not handlers:
                     request.sendall(b"HTTP/1.1 404 NotFound\r\n")
                     log.warning("No Handler")
                     return
-                ctx = Context(self, request, address, log, rfd, method, path, paths, version, hes, route, handlers)
+                ctx = Context(
+                    self, log,
+                    request, rfd, address,
+                    http_method, http_path, http_version,
+                    route, route_keys, http_headers,
+                    register_route, handlers
+                )
                 ctx.next()
                 ctx.flush()
                 if ctx.content_length != len(ctx.content):  # sock should be close
                     return
         except ConnectionAbortedError:
             log.error("connect abort")
+        except ConnectionResetError:
+            log.error("connect reset")
         except:
             log.traceback()
         finally:
@@ -453,6 +513,7 @@ class App:
         self.running = False
 
     def run(self, host="0.0.0.0", port=8080, grace=0, request_queue_size=5, poll_interval=0.5) -> None:
+        Closing.add_close(self.close)
         self.build()
         with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as sock:
             sock.bind((host, port))
@@ -472,7 +533,7 @@ class App:
                         if not ready:
                             continue
                         request, address = sock.accept()
-                        threading.Thread(target=self._, args=(request, address), daemon=True).start()
+                        threading.Thread(target=self.handler_request, args=(request, address), daemon=True).start()
             for i in range(grace):
                 self.log.update(hit=i + 1, grace=grace).warning("server closing")
                 time.sleep(1)
